/*
  CPP_CONTEST=2012
  CPP_PROBLEM=E
  CPP_LANG=C+OPENMP+MPI
  CPP_NUM_CORES=32
  CPP_PROCESSES_PER_NODE=1
 */

/* RECORD
Juan Mompeán Esteban
student of Methodology of  Parallel Programming,
at the University of Murcia, Spain
January 25, 2013
time 2389 msec
speed-up 6.74
*/ 

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>
#include <omp.h>

#define EMPTY 0
#define UP 1
#define RIGHT 2
#define DOWN 3
#define LEFT 4
#define MARK 5 //this value in a position in the checkerboard indicates the position has been marked
#define NUM_PROCESADORES 8

int impulsing(int n,int *cb,int ix,int iy,int *fx,int *fy)
{
	//if the position is empty, store it and indicate a position has been found
	if(ix < 0 || ix >= n || iy < 0 || iy >= n || cb[ix*n+iy] != EMPTY)
	{
		return 0;
	}
	*fx = ix;
	*fy = iy;
	cb[ix * n + iy] = MARK;
	return 1;  
}

int move(int n,int *cb,int ix,int iy,int *fx,int *fy)
{
//if the position is out of the checkerboard do nothing
  if(ix<0 || ix>=n || iy<0 || iy>=n)
  {
    return 0;
  }
  switch(cb[ix*n+iy]) {
  	case EMPTY:
  		*fx = ix;
		*fy = iy;
		cb[(ix * n) + iy] = MARK;
		return 1;
  	case UP:
  		return(impulsing(n,cb,ix,iy-1,fx,fy));
  	case RIGHT:
	  	return(impulsing(n,cb,ix-1,iy,fx,fy));
	case DOWN:
		return(impulsing(n,cb,ix,iy+1,fx,fy));
	case LEFT:
		return(impulsing(n,cb,ix+1,iy,fx,fy));
	default:
		return 0;
  }
}

int minimummov(int n,int *cb,int sx,int sy,int tx,int ty)
{
	int *activesx,*activesy; //coordenates of the actives positions
	int nactives; //number of active positions
	int *tempx,*tempy; //auxiliar vector to store temporally the coordenates of the actives positions
	int *temp,i,j,k;
	int step,end, tempsx, tempsy;

	if(sx == tx && sy == ty) {
		return 0;
	}

	activesx=(int *) malloc(sizeof(int)*n*n);
	activesy=(int *) malloc(sizeof(int)*n*n);
	tempx=(int *) malloc(sizeof(int)*n*n);
	tempy=(int *) malloc(sizeof(int)*n*n);

	cb[sx*n+sy]=MARK; //the source is marked
	activesx[0]=sx;
	activesy[0]=sy;
	nactives=1;
	step=0;
	end=0;
	do
	{
		j = 0;
		for(i = 0; i < nactives; i++)
		{
			for(k = 0; k < 4; k++) 
			{
				tempsx = activesx[i];
				tempsy = activesy[i];
				switch(k) {
					case 0:
						tempsy--;
					break;
					case 1:
						tempsx--;
					break;
					case 2:
						tempsy++;
					break;
					case 3:
						tempsx++;
					break;
				}
				//If movement is posible the position is stored and the number of stored positions is incremented
				if(move(n,cb,tempsx,tempsy,&tempx[j],&tempy[j]))
				{
					if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
					{
						end=1;
						break;
					}
					j++;
				}
			}
		}
		nactives=j;
		temp=activesx;
		activesx=tempx;
		tempx=temp;
		temp=activesy;
		activesy=tempy;
		tempy=temp;
		step++;
	} while(nactives && !end);
	free(activesx);
	free(activesy);
	free(tempx);
	free(tempy);
	return step;
}

void sec(int n,int *cb,int p,int *positions,int *movements,int node,int np)
{
	int i, movement, *cb_aux;
	MPI_Status status;
	if (node != 0) {
		cb = (int *) malloc(sizeof(int) * n * n);
		positions = (int *) malloc(sizeof(int) * 4 * p);
		movements = (int *) malloc(sizeof(int) * p);
	}
	cb_aux = (int *) malloc(sizeof(int) * NUM_PROCESADORES * n * n);
    MPI_Bcast((void*)cb,  n * n, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast((void*)positions, p * 4, MPI_INT, 0, MPI_COMM_WORLD);
	//for each pair of points the minimum number of movements to go from the source to the target is obtained
	omp_set_num_threads(NUM_PROCESADORES);
#pragma omp parallel for private(i) 
	for(i = 0; i < p; i++)
	{	
		if ((i % np) == node) {
#pragma omp critical
{
			memcpy(&cb_aux[omp_get_thread_num() * n * n], cb, sizeof(int) * n * n);
}
			movements[i] = minimummov(n, &cb_aux[omp_get_thread_num() * n * n], positions[i * 4], positions[(i * 4) + 1], positions[(i * 4) + 2], positions[(i * 4) + 3]);
		}
	}
	if (np > 1) {
		for (i = 0; i < p; i++) {
			if ((i % np) == 0) {
				continue;
			}
			if (node == 0) {
				MPI_Recv((void *)&(movements[i]), 1, MPI_INT, i % np, 7, MPI_COMM_WORLD, &status);
			} else if ((i % np) == node) {
				MPI_Send((void *)&(movements[i]), 1, MPI_INT, 0, 7, MPI_COMM_WORLD);
			}
		}
	}
	if (node != 0) {
		free(cb);
		free(positions);
		free(movements);  
	}
	free(cb_aux);
}
